# -*- coding: utf-8 -*-
# cython: language_level=2
# cython: binding=False
# cython: c_string_type=str, c_string_encoding=utf8
# distutils: language = c++
# distutils: depends = libgolang.h
#
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_golang.pyx provides Python interface to libgolang.{h,cpp}.

See _golang.pxd for package overview.
"""

from __future__ import print_function, absolute_import

# init libgolang runtime & friends early
_init_libgolang()
_init_libpyxruntime()

from cpython cimport PyObject, Py_INCREF, Py_DECREF, PY_MAJOR_VERSION
ctypedef PyObject *pPyObject # https://github.com/cython/cython/issues/534
cdef extern from "Python.h":
    ctypedef struct PyTupleObject:
        PyObject **ob_item
    void Py_FatalError(const char *msg)

from libcpp.vector cimport vector
from cython cimport final

import sys

# ---- panic ----

@final
cdef class _PanicError(Exception):
    pass

# panic stops normal execution of current goroutine.
cpdef pypanic(arg):
    raise _PanicError(arg)

# topyexc converts C-level panic/exc to python panic/exc.
# (see usage in e.g. *_pyexc functions in "misc")
cdef void topyexc() except *:
    # TODO use libunwind/libbacktrace/libstacktrace/... to append C-level traceback
    #      from where it panicked till topyexc user.
    # TODO install C-level traceback dump as std::terminate handler.
    #
    # recover_ is declared `except +` - if it was another - not panic -
    # exception, it will be converted to py exc by cython automatically.
    arg = recover_()
    if arg != NULL:
        pyarg = <bytes>arg
        if PY_MAJOR_VERSION >= 3:
            pyarg = pyarg.decode("utf-8")
        pypanic(pyarg)

cdef extern from "golang/libgolang.h" nogil:
    const char *recover_ "golang::recover" () except +


# pyrecover needs to clear sys.exc_info().
# py2 has sys.exc_clear() but it was removed in py3.
# provide our own _pysys_exc_clear().
cdef extern from *:
    """
    static void XPySys_ExcClear() {
    #if PY_VERSION_HEX >= 0x03000000 || defined(PYPY_VERSION)
        PyErr_SetExcInfo(NULL, NULL, NULL);
    #else
        PyThreadState *ts;
        PyObject *exc_type, *exc_value, *exc_traceback;

        ts = PyThreadState_GET();
        exc_type        = ts->exc_type;         ts->exc_type        = NULL;
        exc_value       = ts->exc_value;        ts->exc_value       = NULL;
        exc_traceback   = ts->exc_traceback;    ts->exc_traceback   = NULL;

        Py_XDECREF(exc_type);
        Py_XDECREF(exc_value);
        Py_XDECREF(exc_traceback);
    #endif
    }
    """
    void XPySys_ExcClear()
def _pysys_exc_clear():
    XPySys_ExcClear()

# ---- go ----

# go spawns lightweight thread.
#
# go spawns:
#
# - lightweight thread (with    gevent integration), or
# - full OS thread     (without gevent integration).
#
# Use gpython to run Python with integrated gevent, or use gevent directly to do so.
def pygo(f, *argv, **kw):
    _ = _togo(); _.f = f; _.argv = argv; _.kw    = kw
    Py_INCREF(_)    # we transfer 1 ref to _goviac
    with nogil:
        _taskgo_pyexc(_goviac, <void*>_)

@final
cdef class _togo:
    cdef object f
    cdef tuple  argv
    cdef dict   kw

cdef extern from "Python.h" nogil:
    ctypedef struct PyGILState_STATE:
        pass
    PyGILState_STATE PyGILState_Ensure()
    void PyGILState_Release(PyGILState_STATE)

cdef void _goviac(void *arg) nogil:
    # create new py thread state and keep it alive while __goviac runs.
    #
    # Just `with gil` is not enough: for `with gil` if exceptions could be
    # raised inside, cython generates several GIL release/reacquire calls.
    # This way the thread state will be deleted on first release and _new_ one
    # - _another_ thread state - create on acquire. All that implicitly with
    # the effect of loosing things associated with thread state - e.g. current
    # exception.
    #
    # -> be explicit and manually keep py thread state alive ourselves.
    gstate = PyGILState_Ensure() # py thread state will stay alive until PyGILState_Release
    __goviac(arg)
    PyGILState_Release(gstate)

cdef void __goviac(void *arg) nogil:
    with gil:
        try:
            _ = <_togo>arg
            Py_DECREF(_)
            _.f(*_.argv, **_.kw)
        except:
            # ignore exceptions during python interpreter shutdown.
            # python clears sys and other modules at exit which can result in
            # arbitrary exceptions in still alive "daemon" threads that go
            # spawns. Similarly to threading.py(*) we just ignore them.
            #
            # if we don't - there could lots of output like e.g. "lost sys.stderr"
            # and/or "sys.excepthook is missing" etc.
            #
            # (*) github.com/python/cpython/tree/v2.7.16-121-g53639dd55a0/Lib/threading.py#L760-L778
            #     see also "Technical details" in stackoverflow.com/a/12807285/9456786.
            if sys is None:
                return

            raise   # XXX exception -> exit program with traceback (same as in go) ?


# ---- channels ----

@final
cdef class pychan:
    def __cinit__(pychan pych, size=0, dtype=object):
        pych.dtype = parse_dtype(dtype)
        pych._ch = _makechan_pyexc(dtypeRegistry[<int>pych.dtype].size, size)

    # pychan.nil(X) creates new nil pychan with specified dtype.
    # TODO try to avoid exposing .nil on pychan instances, and expose only pychan.nil
    # http://code.activestate.com/recipes/578486-class-only-methods/
    @staticmethod
    def nil(dtype):
        return pychan._nil(dtype)
    @staticmethod
    cdef pychan _nil(object dtype):
        return pynil(parse_dtype(dtype))

    def __dealloc__(pychan pych):
        if pych._ch == NULL:
            return

        # pychan[X!=object]: just decref the raw chan and we are done.
        if pych.dtype != DTYPE_PYOBJECT:
            _chanxdecref(pych._ch)
            pych._ch = NULL
            return

        # pychan[object], for now, owns the underlying channel.
        # drain buffered channel to decref sent objects.
        # verify that the channel is not connected anywhere outside us.
        # (if it was present also somewhere else - draining would be incorrect)
        #
        # TODO: in the future there could be multiple pychan[object] wrapping
        # the same underlying raw channel: e.g. ch=chan(); ch2=ch.txonly()
        # -> drain underlying channel only when its last reference goes to 0.
        cdef int refcnt = _chanrefcnt(pych._ch)
        if refcnt != 1:
            # cannot raise py-level exception in __dealloc__
            Py_FatalError("pychan[object].__dealloc__: chan.refcnt=%d ; must be =1" % refcnt)

        cdef chan[pPyObject] ch = _wrapchan[pPyObject](pych._ch)
        _chanxdecref(pych._ch)
        pych._ch = NULL

        cdef PyObject *_rx
        while ch.len() != 0:
            # NOTE *not* chanrecv_pyexc(ch):
            # - recv must not block and must not panic as we verified that we
            #   are the only holder of the channel and that ch buffer is not empty.
            # - even if recv panics, we cannot convert that panic to python
            #   exception in __dealloc__. So if it really panics - let the
            #   panic make it and crash the process similarly to Py_FatalError above.
            _rx = ch.recv()
            Py_DECREF(<object>_rx)

        # ch is decref'ed automatically at return


    # send sends object to a receiver.
    def send(pychan pych, obj):
        cdef chanElemBuf _tx = 0

        if pych.dtype == DTYPE_PYOBJECT:
            # increment obj reference count - until received the channel is
            # holding pointer to the object.
            Py_INCREF(obj)
            (<PyObject **>&_tx)[0] = <PyObject *>obj
        else:
            py_to_c(pych.dtype, obj, &_tx)

        try:
            with nogil:
                _chansend_pyexc(pych._ch, &_tx)
        except: # not only _PanicError as send can also throw e.g. bad_alloc
            # the object was not sent - e.g. it was "send on a closed channel"
            if pych.dtype == DTYPE_PYOBJECT:
                Py_DECREF(obj)
            raise

    # recv_ is "comma-ok" version of recv.
    #
    # ok is true - if receive was delivered by a successful send.
    # ok is false - if receive is due to channel being closed and empty.
    def recv_(pychan pych): # -> (rx, ok)
        cdef chanElemBuf _rx = 0
        cdef bint ok

        with nogil:
            ok = _chanrecv__pyexc(pych._ch, &_rx)

        cdef object rx = None
        cdef PyObject *_rxpy
        if pych.dtype == DTYPE_PYOBJECT:
            _rxpy = (<PyObject **>&_rx)[0]
            if _rxpy != NULL:
                # we received the object and the channel dropped pointer to it.
                rx = <object>_rxpy
                Py_DECREF(rx)
        else:
            rx = c_to_py(pych.dtype, &_rx)

        return (rx, ok)

    # recv receives from the channel.
    def recv(pychan pych): # -> rx
        rx, _ = pych.recv_()    # TODO call recv_ via C
        return rx

    # close closes sending side of the channel.
    def close(pychan pych):
        with nogil:
            _chanclose_pyexc(pych._ch)

    def __len__(pychan pych):
        return _chanlen_pyexc(pych._ch)

    def __repr__(pychan pych):
        if pych._ch == NULL:
            if pych.dtype == DTYPE_PYOBJECT:
                return "nilchan"
            else:
                return "chan.nil(%r)" % dtypeinfo(pych.dtype).name
        else:
            return super(pychan, pych).__repr__()

    # pychan == pychan
    def __hash__(pychan pych):
        return <Py_hash_t>pych._ch
    def __eq__(pychan a, object rhs):
        if not isinstance(rhs, pychan):
            return False
        cdef pychan b = rhs
        if a._ch != b._ch:
            return False
        # a and b point to the same underlying channel object.
        # they compare as same if their types match, or, for nil, additionally,
        # if one of the sides is nil[*] (untyped nil).
        if a.dtype == b.dtype:
            return True
        if a._ch != NULL:
            return False
        if a.dtype == DTYPE_PYOBJECT or b.dtype == DTYPE_PYOBJECT:
            return True
        return False


    # pychan -> chan[X]
    # ( should live in "runtime support for channel types" if we could define
    #   methods separate from class )
    cdef nogil:
        chan[structZ]   chan_structZ    (pychan pych):
            pychan_asserttype(pych, DTYPE_STRUCTZ)
            return _wrapchan[structZ](pych._ch)

        chan[cbool]     chan_bool       (pychan pych):
            pychan_asserttype(pych, DTYPE_BOOL)
            return _wrapchan[cbool] (pych._ch)

        chan[int]       chan_int        (pychan pych):
            pychan_asserttype(pych, DTYPE_INT)
            return _wrapchan[int]    (pych._ch)

        chan[double]    chan_double     (pychan pych):
            pychan_asserttype(pych, DTYPE_DOUBLE)
            return _wrapchan[double] (pych._ch)

    # pychan <- chan[X]
    @staticmethod
    cdef pychan from_chan_structZ   (chan[structZ] ch):
        return pychan_from_raw(ch._rawchan(),   DTYPE_STRUCTZ)

    @staticmethod
    cdef pychan from_chan_bool      (chan[cbool] ch):
        return pychan_from_raw(ch._rawchan(),   DTYPE_BOOL)

    @staticmethod
    cdef pychan from_chan_int       (chan[int] ch):
        return pychan_from_raw(ch._rawchan(),   DTYPE_INT)

    @staticmethod
    cdef pychan from_chan_double    (chan[double] ch):
        return pychan_from_raw(ch._rawchan(),   DTYPE_DOUBLE)

cdef void pychan_asserttype(pychan pych, DType dtype) nogil:
    if pych.dtype != dtype:
        panic("pychan: channel type mismatch")

cdef pychan pychan_from_raw(_chan *_ch, DType dtype):
    cdef pychan pych = pychan.__new__(pychan)
    pych.dtype = dtype
    pych._ch   = _ch; _chanxincref(_ch)
    return pych


# pydefault represents default case for pyselect.
pydefault  = object()

# pyselect executes one ready send or receive channel case.
#
# if no case is ready and default case was provided, select chooses default.
# if no case is ready and default was not provided, select blocks until one case becomes ready.
#
# returns: selected case number and receive info (None if send case was selected).
#
# example:
#
#   _, _rx = select(
#       ch1.recv,           # 0
#       ch2.recv_,          # 1
#       (ch2.send, obj2),   # 2
#       default,            # 3
#   )
#   if _ == 0:
#       # _rx is what was received from ch1
#       ...
#   if _ == 1:
#       # _rx is (rx, ok) of what was received from ch2
#       ...
#   if _ == 2:
#       # we know obj2 was sent to ch2
#       ...
#   if _ == 3:
#       # default case
#       ...
def pyselect(*pycasev):
    cdef int i, n = len(pycasev), selected
    cdef vector[_selcase] casev = vector[_selcase](n, default)
    cdef pychan pych
    cdef chanElemBuf _rx = 0  # all select recvs are setup to receive into _rx
    cdef cbool rxok = False   # (its ok as only one receive will be actually executed)

    selected = -1
    try:
        # prepare casev for chanselect
        for i in range(n):
            pycase = pycasev[i]
            # default
            if pycase is pydefault:
                casev[i] = default

            # send
            elif type(pycase) is tuple:
                if len(pycase) != 2:
                    pypanic("pyselect: invalid [%d]() case" % len(pycase))
                _tcase = <PyTupleObject *>pycase

                pysend = <object>(_tcase.ob_item[0])
                if pysend.__self__.__class__ is not pychan:
                    pypanic("pyselect: send on non-chan: %r" % (pysend.__self__.__class__,))
                pych = pysend.__self__

                if pysend.__name__ != "send":       # XXX better check PyCFunction directly
                    pypanic("pyselect: send expected: %r" % (pysend,))

                tx = <object>(_tcase.ob_item[1])
                casev[i] = _selsend(pych._ch, NULL)
                casev[i].flags = _INPLACE_DATA
                casev[i].user  = pych.dtype

                if pych.dtype == DTYPE_PYOBJECT:
                    # incref tx as if corresponding channel is holding pointer to the object while it is being sent.
                    # we'll decref the object if it won't be sent.
                    # see pychan.send for details.
                    Py_INCREF(tx)
                    (<PyObject **>&casev[i].itxrx)[0] = <PyObject *>tx

                else:
                    # NOTE vvv assumes that size for all dtypes fit into 64 bits
                    py_to_c(pych.dtype, tx, &casev[i].itxrx)  # NOTE can raise exception

            # recv
            else:
                pyrecv = pycase
                if pyrecv.__self__.__class__ is not pychan:
                    pypanic("pyselect: recv on non-chan: %r" % (pyrecv.__self__.__class__,))
                pych = pyrecv.__self__

                if pyrecv.__name__ == "recv":       # XXX better check PyCFunction directly
                    casev[i] = _selrecv(pych._ch, &_rx)
                elif pyrecv.__name__ == "recv_":    # XXX better check PyCFunction directly
                    casev[i] = _selrecv_(pych._ch, &_rx, &rxok)
                else:
                    pypanic("pyselect: recv expected: %r" % (pyrecv,))

                casev[i].user = pych.dtype

        with nogil:
            selected = _chanselect_pyexc(&casev[0], casev.size())

    finally:
        # decref not sent tx (see ^^^ send prepare)
        for i in range(n):
            if casev[i].op == _CHANSEND and casev[i].user == DTYPE_PYOBJECT and (i != selected):
                _tx = (<PyObject **>casev[i].ptx())[0]
                tx  = <object>_tx
                Py_DECREF(tx)

    # return what was selected
    cdef _chanop op = casev[selected].op
    if op == _DEFAULT:
        return selected, None
    if op == _CHANSEND:
        return selected, None

    if op != _CHANRECV:
        raise AssertionError("pyselect: chanselect returned with bad op")

    cdef object rx = None
    cdef PyObject *_rxpy
    cdef DType rxtype = <DType>casev[selected].user
    if rxtype == DTYPE_PYOBJECT:
        # we received NULL or the object; if it is object, corresponding channel
        # dropped pointer to it (see pychan.recv_ for details).
        _rxpy = (<PyObject **>&_rx)[0]
        if _rxpy != NULL:
            rx = <object>_rxpy
            Py_DECREF(rx)

    else:
        rx = c_to_py(rxtype, &_rx)

    if casev[selected].rxok != NULL:
        return selected, (rx, rxok)
    else:
        return selected, rx

# ---- init libgolang runtime ---

cdef extern from "golang/libgolang.h" namespace "golang" nogil:
    struct _libgolang_runtime_ops
    void _libgolang_init(const _libgolang_runtime_ops*)
from cpython cimport PyCapsule_Import

cdef void _init_libgolang() except*:
    # detect whether we are running under gevent or OS threads mode
    # -> use golang.runtime._runtime_(gevent|thread) as libgolang runtime.
    threadmod = "thread"
    if PY_MAJOR_VERSION >= 3:
        threadmod = "_thread"
    t = __import__(threadmod)
    runtime = "thread"
    if "gevent" in t.start_new_thread.__module__:
        runtime = "gevent"
    runtimemod = "golang.runtime." + "_runtime_" + runtime

    # PyCapsule_Import("golang.X") does not work properly while we are in the
    # process of importing golang (it tries to access "X" attribute of half-created
    # golang module). -> preimport runtimemod via regular import first.
    __import__(runtimemod)
    runtimecaps = (runtimemod + ".libgolang_runtime_ops").encode("utf-8") # py3
    cdef const _libgolang_runtime_ops *runtime_ops = \
        <const _libgolang_runtime_ops*>PyCapsule_Import(runtimecaps, 0)
    if runtime_ops == NULL:
        pypanic("init: %s: libgolang_runtime_ops=NULL" % runtimemod)
    _libgolang_init(runtime_ops)


cdef void _init_libpyxruntime() except*:
    # this initializes libpyxruntime and registers its pyatexit hook
    import golang.pyx.runtime


# ---- misc ----

cdef extern from "golang/libgolang.h" namespace "golang" nogil:
    _chan  *_makechan(unsigned elemsize, unsigned size)
    chan[T] _wrapchan[T](_chan *)
    void    _chanxincref(_chan *ch)
    void    _chanxdecref(_chan *ch)
    int     _chanrefcnt(_chan *ch)
    void    _chansend(_chan *ch, const void *ptx)
    bint    _chanrecv_(_chan *ch, void *prx)
    void    _chanclose(_chan *ch)
    unsigned _chanlen(_chan *ch)

    int  _chanselect(_selcase *casev, int casec)
    _selcase _selsend(_chan *ch, const void *ptx)
    _selcase _selrecv(_chan *ch, void *prx)
    _selcase _selrecv_(_chan *ch, void *prx, bint *pok)

    void _taskgo(void (*f)(void *), void *arg)

cdef nogil:

    _chan* _makechan_pyexc(unsigned elemsize, unsigned size)    except +topyexc:
        return _makechan(elemsize, size)

    void _chansend_pyexc(_chan *ch, const void *ptx)            except +topyexc:
        _chansend(ch, ptx)

    bint _chanrecv__pyexc(_chan *ch, void *prx)                 except +topyexc:
        return _chanrecv_(ch, prx)

    void _chanclose_pyexc(_chan *ch)                            except +topyexc:
        _chanclose(ch)

    unsigned _chanlen_pyexc(_chan *ch)                          except +topyexc:
        return _chanlen(ch)

    int _chanselect_pyexc(const _selcase *casev, int casec)     except +topyexc:
        return _chanselect(casev, casec)

    void _taskgo_pyexc(void (*f)(void *) nogil, void *arg)      except +topyexc:
        _taskgo(f, arg)


# ---- runtime support for channel types ----

# chanElemBuf is large enough to keep any dtype.
# NOTE sizeof(chanElemBuf) = max(_.size) from dtypeRegistry (this is checked at
# runtime on module init).
ctypedef uint64_t chanElemBuf

# DTypeInfo provides runtime information for a DType:
# dtype name, element size, py <-> c element conversion routines and typed nil instance.
cdef struct DTypeInfo:
    const char *name    # name of the type, e.g. "C.int"
    unsigned    size

    # py_to_c converts python object to C-level data of dtype.
    # If conversion fails, corresponding exception is raised.
    bint      (*py_to_c)(object obj, chanElemBuf *cto) except False

    # c_to_py converts C-level data into python object according to dtype.
    # The conversion cannot fail, but this can raise exception e.g. due to
    # error when allocating result object.
    object    (*c_to_py) (const chanElemBuf *cfrom)

    # pynil points to pychan instance that represents nil[dtype].
    # it holds one reference and is never freed.
    PyObject   *pynil

# py_to_c converts Python object to C-level data according to dtype.
cdef bint py_to_c(DType dtype, object obj, chanElemBuf *cto) except False:
    dtypei = dtypeinfo(dtype)
    return dtypei.py_to_c(obj, cto)

# c_to_py converts C-level data into Python object according to dtype.
cdef object c_to_py(DType dtype, const chanElemBuf *cfrom):
    dtypei = dtypeinfo(dtype)
    return dtypei.c_to_py(cfrom)

# mkpynil creates pychan instance that represents nil[dtype].
cdef PyObject *mkpynil(DType dtype):
    cdef pychan pynil = pychan.__new__(pychan)
    pynil.dtype = dtype
    pynil._ch   = NULL   # should be already NULL
    Py_INCREF(pynil)
    return <PyObject *>pynil

# pynil returns pychan instance corresponding to nil[dtype].
cdef pychan pynil(DType dtype):
    dtypei = dtypeinfo(dtype)
    return <pychan>dtypei.pynil

# {} dtype -> DTypeInfo.
# XXX const
cdef DTypeInfo[<int>DTYPE_NTYPES] dtypeRegistry

# dtypeinfo returns DTypeInfo corresponding to dtype.
cdef DTypeInfo* dtypeinfo(DType dtype) nogil:
    if not (0 <= dtype < DTYPE_NTYPES):
        # no need to ->pyexc, as this bug means memory corruption and so is fatal
        panic("BUG: pychan dtype invalid")
    return &dtypeRegistry[<int>dtype]

# DTYPE_PYOBJECT
dtypeRegistry[<int>DTYPE_PYOBJECT] = DTypeInfo(
    name        = "object",
    size        = sizeof(PyObject*),
    # py_to_c/c_to_py must not be called for pyobj - as pyobj is the common
    # case, they are manually inlined for speed. The code is also more clear
    # when Py_INCREF/Py_DECREF go in send/recv/select directly.
    py_to_c     = NULL, # must not be called for pyobj
    c_to_py     = NULL, # must not be called for pyobj
    pynil       = mkpynil(DTYPE_PYOBJECT),
)

# pynilchan is nil py channel.
#
# On nil channel: send/recv block forever; close panics.
# pynilchan is alias for pychan.nil(object).
pynilchan = pychan._nil(object)


# DTYPE_STRUCTZ
dtypeRegistry[<int>DTYPE_STRUCTZ] = DTypeInfo(
    name        = "C.structZ",
    size        = 0, # NOTE = _elemsize<structZ>, but _not_ sizeof(structZ) which = 1
    py_to_c     = structZ_py_to_c,
    c_to_py     = structZ_c_to_py,
    pynil       = mkpynil(DTYPE_STRUCTZ),
)

cdef bint structZ_py_to_c(object obj, chanElemBuf *cto) except False:
    # for structZ the only accepted value from python is None
    if obj is not None:
        raise TypeError("type mismatch: expect structZ; got %r" % (obj,))
    # nothing to do - size = 0
    return True

cdef object structZ_c_to_py(const chanElemBuf *cfrom):
    return None


# DTYPE_BOOL
dtypeRegistry[<int>DTYPE_BOOL] = DTypeInfo(
    name        = "C.bool",
    size        = sizeof(cbool),
    py_to_c     = bool_py_to_c,
    c_to_py     = bool_c_to_py,
    pynil       = mkpynil(DTYPE_BOOL),
)

cdef bint bool_py_to_c(object obj, chanElemBuf *cto) except False:
    # don't accept int/double/str/whatever.
    if type(obj) is not bool:
        raise TypeError("type mismatch: expect bool; got %r" % (obj,))
    (<cbool *>cto)[0] = obj # raises *Error if conversion fails
    return True

cdef object bool_c_to_py(const chanElemBuf *cfrom):
    return (<cbool *>cfrom)[0]


# DTYPE_INT
dtypeRegistry[<int>DTYPE_INT] = DTypeInfo(
    name        = "C.int",
    size        = sizeof(int),
    py_to_c     = int_py_to_c,
    c_to_py     = int_c_to_py,
    pynil       = mkpynil(DTYPE_INT),
)

cdef bint int_py_to_c(object obj, chanElemBuf *cto) except False:
    # don't accept bool
    if isinstance(obj, bool):
        raise TypeError("type mismatch: expect int; got %r" % (obj,))
    # don't allow e.g. 3.14 to be implicitly truncated to just 3
    cdef double objf = obj
    if (<int>objf) != objf:
        raise TypeError("type mismatch: expect int; got %r" % (obj,))
    (<int *>cto)[0] = obj # raises *Error if conversion fails
    return True

cdef object int_c_to_py(const chanElemBuf *cfrom):
    return (<int *>cfrom)[0]


# DTYPE_DOUBLE
dtypeRegistry[<int>DTYPE_DOUBLE] = DTypeInfo(
    name        = "C.double",
    size        = sizeof(double),
    py_to_c     = double_py_to_c,
    c_to_py     = double_c_to_py,
    pynil       = mkpynil(DTYPE_DOUBLE),
)

cdef bint double_py_to_c(object obj, chanElemBuf *cto) except False:
    # don't accept bool
    if isinstance(obj, bool):
        raise TypeError("type mismatch: expect float; got %r" % (obj,))
    (<double *>cto)[0] = obj # raises *Error if conversion fails
    return True

cdef object double_c_to_py(const chanElemBuf *cfrom):
    return (<double *>cfrom)[0]


# verify at init time that sizeof(chanElemBuf) = max(_.size)
cdef verify_chanElemBuf():
    cdef int size_max = 0
    for dtype in range(DTYPE_NTYPES):
        size_max = max(size_max, dtypeRegistry[<int>dtype].size)
    if size_max != sizeof(chanElemBuf):
        raise AssertionError("golang: module is miscompiled: max(dtype.size) = %d  ; compiled with = %d" % (size_max, sizeof(chanElemBuf)))
verify_chanElemBuf()

# {} dtype name (str) -> dtype
cdef dict name2dtype = {}
cdef init_name2dtype():
    for dtype in range(DTYPE_NTYPES):
        name2dtype[dtypeRegistry[<int>dtype].name] = dtype
init_name2dtype()

# parse_dtype converts object or string dtype, as e.g. passed to
# pychan(dtype=...) into DType.
cdef DType parse_dtype(dtype) except <DType>-1:
    if dtype is object:
        return DTYPE_PYOBJECT

    _ = name2dtype.get(dtype)
    if _ is None:
        raise TypeError("pychan: invalid dtype: %r" % (dtype,))
    return _



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package context mirrors and amends Go package context.

 - `Context` represents operational context that carries deadline, cancellation
   signal and immutable context-local key -> value dict.
 - `background` returns empty context that is never canceled.
 - `with_cancel` creates new context that can be canceled on its own.
 - `with_deadline` creates new context with deadline.
 - `with_timeout` creates new context with timeout.
 - `with_value` creates new context with attached key=value.
 - `merge` creates new context from 2 parents(*).

See also https://golang.org/pkg/context for Go context package documentation.
See also https://blog.golang.org/context for overview.

(*) not provided in Go version.
"""

from golang cimport chan, structZ, error, refptr, interface
from golang cimport cxx
from libcpp.utility cimport pair

# XXX for std::function cython does not provide operator() and =nullptr
#from libcpp.functional cimport function
#ctypedef function[void()] cancelFunc
cdef extern from "golang/libgolang.h" namespace "golang" nogil:
    cppclass cancelFunc "golang::func<void()>":
        void operator() ()
        void operator= (nullptr_t)

cdef extern from "golang/context.h" namespace "golang::context" nogil:
    cppclass _Context:
        double          deadline()
        chan[structZ]   done()
        error           err()
        interface       value(const void *key)

    cppclass Context (refptr[_Context]):
        # Context.X = Context->X in C++
        double          deadline    "_ptr()->deadline"  ()
        chan[structZ]   done        "_ptr()->done"      ()
        error           err         "_ptr()->err"       ()
        interface       value       "_ptr()->value"     (const void *key)

    Context background()
    error   canceled
    error   deadlineExceeded

    pair[Context, cancelFunc] with_cancel   (Context parent)
    Context                   with_value    (Context parent, const void *key, interface value)
    pair[Context, cancelFunc] with_deadline (Context parent, double deadline)
    pair[Context, cancelFunc] with_timeout  (Context parent, double timeout)
    pair[Context, cancelFunc] merge         (Context parent1, Context parent2)

    # for testing
    cxx.set[Context] _tctxchildren(Context ctx)


# ---- python bits ----

from golang cimport pychan
from cython cimport final

@final
cdef class PyContext:
    cdef Context  ctx
    cdef pychan   _pydone # pychan wrapping ctx.done()

    # PyContext.from_ctx returns PyContext wrapping pyx/nogil-level Context ctx.
    @staticmethod
    cdef PyContext from_ctx (Context ctx)



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package time mirrors Go package time.

 - `now` returns current time.
 - `sleep` pauses current task.
 - `Ticker` and `Timer` provide timers integrated with channels.
 - `tick`, `after` and `after_func` are convenience wrappers to use
   tickers and timers easily.

See also https://golang.org/pkg/time for Go time package documentation.
"""

from golang cimport chan, cbool, refptr
from libcpp cimport nullptr_t

# golang/pyx - the same as std python - represents time as float
cdef extern from * nogil:
    # XXX how to declare/share constants without C verbatim?
    """
    #ifndef _golang_time_pxd_h
    #define _golang_time_pxd_h
    # define    golang_time_second      (1.0)
    # define    golang_time_nanosecond  (1E-9 * golang_time_second)
    # define    golang_time_microsecond (1E-6 * golang_time_second)
    # define    golang_time_millisecond (1E-3 * golang_time_second)
    # define    golang_time_minute      (60   * golang_time_second)
    # define    golang_time_hour        (60   * golang_time_minute)
    #endif // _golang_time_pxd_h
    """
    const double second         "golang_time_second"
    const double nanosecond     "golang_time_nanosecond"
    const double microsecond    "golang_time_microsecond"
    const double millisecond    "golang_time_millisecond"
    const double minute         "golang_time_minute"
    const double hour           "golang_time_hour"


cdef extern from "golang/time.h" namespace "golang::time" nogil:
    void   sleep(double dt)
    double now()

    chan[double] tick(double dt)
    chan[double] after(double dt)
    Timer        after_func(double dt, ...)    # ... = func<void()>

    cppclass _Ticker:
        chan[double] c
        void stop()

    cppclass Ticker (refptr[_Ticker]):
        # Ticker.X = Ticker->X in C++.
        chan[double] c      "_ptr()->c"
        void         stop   "_ptr()->stop" ()

    Ticker new_ticker(double dt)


    cppclass _Timer:
        chan[double] c
        cbool stop()
        void  reset(double dt)

    cppclass Timer (refptr[_Timer]):
        # Timer.X = Timer->X in C++.
        chan[double] c      "_ptr()->c"
        cbool        stop   "_ptr()->stop"  ()
        void         reset  "_ptr()->reset" (double dt)

    Timer new_timer(double dt)



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""pytest: py2: pygolang integration patches."""

from __future__ import print_function, absolute_import

from golang import _patch
import inspect

# _PY2 adds support for chained exceptions created by defer.
#
# It returns True by default (we are running under py2) except when called from
# _pytest/_code/code.FormattedExcinfo.repr_excinfo() for which it pretends to
# be running py3 if raised exception has .__cause__ .
class _PY2:
    @staticmethod
    def __nonzero__():
        fcall = inspect.currentframe().f_back
        if fcall.f_code.co_name != "repr_excinfo":
            return True     # XXX also check class/module?
        exci = fcall.f_locals.get('excinfo', None)
        if exci is None:
            return True
        exc = getattr(exci, 'value', None)
        if exc is None:
            return True
        if not hasattr(exc, '__cause__'):
            return True
        return False

@_patch.afterimport('_pytest')
def _():
    from _pytest._code import code
    code._PY2 = _PY2()



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang.golang_test import pyrun, pyout
from os.path import dirname

testprog = dirname(__file__) + "/testprog"

# verify that we can build/run external package that uses pygolang in pyx mode.
def test_pyx_build():
    pyxuser = testprog + "/golang_pyx_user"
    pyrun(["setup.py", "build_ext", "-i"], cwd=pyxuser)

    # run built test.
    _ = pyout(["-c",
        # XXX `import golang` is a hack: it dynamically loads _golang.so -> libgolang.so,
        # and this way dynamic linker already has libgolang.so DSO found and in
        # memory for all further imports. If we don't, current state of setuptools_dso
        # is that `import pyxuser.test` will fail finding libgolang.so.
        "import golang;" +
        "from pyxuser import test; test.main()"], cwd=pyxuser)
    assert _ == b"test.pyx: OK\n"


# verify that custom classes can be used via cmdclass
def test_pyx_build_cmdclass():
    _ = pyout(["cmdclass_custom.py", "build_ext"], cwd=testprog)
    assert b"pyx.build:RUN_BUILD_EXT" in _



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_time.pyx implements time.pyx - see _time.pxd for package overview."""

from __future__ import print_function, absolute_import

from golang cimport pychan, topyexc
from golang cimport sync
from golang.pyx cimport runtime
from cpython cimport PyObject
from cython cimport final


def pynow(): # -> t
    return now_pyexc()

def pysleep(double dt):
    with nogil:
        sleep_pyexc(dt)


# ---- timers ----

# tick returns channel connected to dt ticker.
#
# Note: there is no way to stop created ticker.
# Note: for dt <= 0, contrary to Ticker, tick returns nil channel instead of panicking.
def pytick(double dt):  # -> chan time
    return pychan.from_chan_double( tick(dt) )

# after returns channel connected to dt timer.
#
# Note: with after there is no way to stop/garbage-collect created timer until it fires.
def pyafter(double dt): # -> chan time
    return pychan.from_chan_double( after(dt) )

# after_func arranges to call f after dt time.
#
# The function will be called in its own goroutine.
# Returned timer can be used to cancel the call.
def pyafter_func(double dt, f):  # -> PyTimer
    return PyTimer(dt, f=f)


# Ticker arranges for time events to be sent to .c channel on dt-interval basis.
#
# If the receiver is slow, Ticker does not queue events and skips them.
# Ticking can be canceled via .stop() .
@final
cdef class PyTicker:
    cdef Ticker   tx
    cdef readonly pychan  c # pychan wrapping tx.c

    def __init__(PyTicker pytx, double dt):
        with nogil:
            pytx.tx = new_ticker_pyexc(dt)
        pytx.c = pychan.from_chan_double( pytx.tx.c )

    def __dealloc__(PyTicker pytx):
        pytx.tx = NULL

    # stop cancels the ticker.
    #
    # It is guaranteed that ticker channel is empty after stop completes.
    def stop(PyTicker pytx):
        with nogil:
            ticker_stop_pyexc(pytx.tx)


# Timer arranges for time event to be sent to .c channel after dt time.
#
# The timer can be stopped (.stop), or reinitialized to another time (.reset).
#
# If func f is provided - when the timer fires f is called in its own goroutine
# instead of event being sent to channel .c .
@final
cdef class PyTimer:
    cdef Timer    t
    cdef readonly pychan  c # pychan wrapping t.c

    def __init__(PyTimer pyt, double dt, f=None):
        with nogil:
            if f is None:
                pyt.t = new_timer_pyexc(dt)
            else:
                pyt.t = _new_timer_pyfunc_pyexc(dt, <PyObject *>f)
        pyt.c = pychan.from_chan_double( pyt.t.c )

    def __dealloc__(PyTimer pyt):
        pyt.t = NULL

    # stop cancels the timer.
    #
    # It returns:
    #
    #   False: the timer was already expired or stopped,
    #   True:  the timer was armed and canceled by this stop call.
    #
    # Note: contrary to Go version, there is no need to drain timer channel
    # after stop call - it is guaranteed that after stop the channel is empty.
    #
    # Note: similarly to Go, if Timer is used with function - it is not
    # guaranteed that after stop the function is not running - in such case
    # the caller must explicitly synchronize with that function to complete.
    def stop(PyTimer pyt): # -> canceled
        with nogil:
            canceled = timer_stop_pyexc(pyt.t)
        return canceled

    # reset rearms the timer.
    #
    # the timer must be either already stopped or expired.
    def reset(PyTimer pyt, double dt):
        with nogil:
            timer_reset_pyexc(pyt.t, dt)


# ---- misc ----
pysecond        = second
pynanosecond    = nanosecond
pymicrosecond   = microsecond
pymillisecond   = millisecond
pyminute        = minute
pyhour          = hour

cdef nogil:

    double now_pyexc()                                      except +topyexc:
        return now()
    void sleep_pyexc(double dt)                             except +topyexc:
        sleep(dt)

    Ticker new_ticker_pyexc(double dt)                      except +topyexc:
        return new_ticker(dt)
    void ticker_stop_pyexc(Ticker tx)                       except +topyexc:
        tx.stop()
    Timer new_timer_pyexc(double dt)                        except +topyexc:
        return new_timer(dt)
    Timer _new_timer_pyfunc_pyexc(double dt, PyObject *pyf) except +topyexc:
        # NOTE C++ implicitly casts func<void()> <- func<error()>
        # XXX  error (= Py Exception) -> exit program with traceback (same as in go) ?
        return after_func(dt, runtime.PyFunc(pyf))

    cbool timer_stop_pyexc(Timer t)                         except +topyexc:
        return t.stop()
    void timer_reset_pyexc(Timer t, double dt)              except +topyexc:
        t.reset(dt)



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_runtime_thread.pyx provides libgolang runtime based on OS threads."""

from __future__ import print_function, absolute_import

# Thread runtime reuses C-level Python threadcreate + semaphore implementation
# for portability. In Python semaphores do not depend on GIL and by reusing
# the implementation we can offload us from covering different systems.
#
# On POSIX, for example, Python uses sem_init(process-private) + sem_post/sem_wait.
#
# Similarly PyThread_start_new_thread - Python's C function function to create
# new thread - does not depend on GIL. On POSIX, for example, it is small
# wrapper around pthread_create.
#
# NOTE Cython declares PyThread_acquire_lock/PyThread_release_lock as nogil
from cpython.pythread cimport PyThread_acquire_lock, PyThread_release_lock, \
        PyThread_type_lock, WAIT_LOCK

# NOTE On Darwin, even though this is considered as POSIX, Python uses
# mutex+condition variable to implement its lock, and, as of 20190828, Py2.7
# implementation, even though similar issue was fixed for Py3 in 2012, contains
# synchronization bug: the condition is signalled after mutex unlock while the
# correct protocol is to signal condition from under mutex:
#
#   https://github.com/python/cpython/blob/v2.7.16-127-g0229b56d8c0/Python/thread_pthread.h#L486-L506
#   https://github.com/python/cpython/commit/187aa545165d (py3 fix)
#
# PyPy has the same bug for both pypy2 and pypy3:
#
#   https://bitbucket.org/pypy/pypy/src/578667b3fef9/rpython/translator/c/src/thread_pthread.c#lines-443:465
#   https://bitbucket.org/pypy/pypy/src/5b42890d48c3/rpython/translator/c/src/thread_pthread.c#lines-443:465
#
# This way when Pygolang is used with buggy Python/darwin, the bug leads to
# frequently appearing deadlocks, while e.g. CPython3/darwin works ok.
#
# The bug was reported to CPython/PyPy upstreams:
#
# - https://bugs.python.org/issue38106
# - https://bitbucket.org/pypy/pypy/issues/3072
#
# and fixed in CPython 2.7.17 and PyPy 7.2 .
import sys, platform
if 'darwin' in sys.platform:
    pyimpl = platform.python_implementation()
    pyver  = sys.version_info
    buggy  = buglink = None
    if 'CPython' in pyimpl and pyver < (3, 0) and pyver < (2,7,17):
        buggy   = "cpython2/darwin < 2.7.17"
        buglink = "https://bugs.python.org/issue38106"
    if 'PyPy' in pyimpl and sys.pypy_version_info < (7,2):
        buggy   = "pypy/darwin < 7.2"
        buglink = "https://bitbucket.org/pypy/pypy/issues/3072"
    if buggy:
        print("WARNING: pyxgo: thread: %s has race condition bug in runtime"
              " that leads to deadlocks (%s)" % (buggy, buglink), file=sys.stderr)

# make sure python threading is initialized, so that there is no concurrent
# calls to PyThread_init_thread from e.g. PyThread_allocate_lock later.
#
# This allows us to treat PyThread_allocate_lock & PyThread_start_new_thread as nogil.
from cpython.ceval cimport PyEval_InitThreads
#PyThread_init_thread()     # initializes only threading, but _not_ GIL
PyEval_InitThreads()        # initializes      threading       and  GIL
cdef extern from "pythread.h" nogil:
    # NOTE py3.7 changed to `unsigned long PyThread_start_new_thread ...`
    long PyThread_start_new_thread(void (*)(void *), void *)
    PyThread_type_lock PyThread_allocate_lock()
    void PyThread_free_lock(PyThread_type_lock)

from golang.runtime._libgolang cimport _libgolang_runtime_ops, _libgolang_sema, \
        _libgolang_runtime_flags, panic

from libc.stdint cimport uint64_t, UINT64_MAX
IF POSIX:
    from posix.time cimport clock_gettime, nanosleep as posix_nanosleep, timespec, CLOCK_REALTIME
    from libc.errno cimport errno, EINTR
ELSE:
    # !posix via-gil timing fallback
    import time as pytimemod
    from golang.runtime._runtime_pymisc cimport PyExc, pyexc_fetch, pyexc_restore

    cdef:
        bint _nanosleep(double dt_s):
            pytimemod.sleep(dt_s)
            return True

        (double, bint) _nanotime():
            cdef double t_s
            t_s = pytimemod.time()
            return t_s, True


DEF i1E9 = 1000000000
#           987654321

cdef nogil:

    void go(void (*f)(void *), void *arg):
        pytid = PyThread_start_new_thread(f, arg)
        if pytid == -1:
            panic("pygo: failed")

    _libgolang_sema* sema_alloc():
        # python calls it "lock", but it is actually a semaphore.
        # and in particular can be released by thread different from thread that acquired it.
        pysema = PyThread_allocate_lock()
        return <_libgolang_sema *>pysema # NULL is ok - libgolang expects it

    void sema_free(_libgolang_sema *gsema):
        pysema = <PyThread_type_lock>gsema
        PyThread_free_lock(pysema)

    void sema_acquire(_libgolang_sema *gsema):
        pysema = <PyThread_type_lock>gsema
        ok = PyThread_acquire_lock(pysema, WAIT_LOCK)
        if ok == 0:
            panic("pyxgo: thread: sema_acquire: PyThread_acquire_lock failed")

    void sema_release(_libgolang_sema *gsema):
        pysema = <PyThread_type_lock>gsema
        PyThread_release_lock(pysema)

    IF POSIX:
        void nanosleep(uint64_t dt):
            cdef timespec ts
            ts.tv_sec  = dt // i1E9
            ts.tv_nsec = dt  % i1E9
            err = posix_nanosleep(&ts, NULL)
            if err == -1 and errno == EINTR:
                err = 0 # XXX ok?
            if err == -1:
                panic("pyxgo: thread: nanosleep: nanosleep failed") # XXX +errno
    ELSE:
        void nanosleep(uint64_t dt):
            cdef double dt_s = dt * 1E-9 # no overflow possible
            cdef PyExc exc
            with gil:
                pyexc_fetch(&exc)
                ok = _nanosleep(dt_s)
                pyexc_restore(exc)
            if not ok:
                panic("pyxgo: thread: nanosleep: pytime.sleep failed")

    IF POSIX:
        uint64_t nanotime():
            cdef timespec ts
            cdef int err = clock_gettime(CLOCK_REALTIME, &ts)
            if err == -1:
                panic("pyxgo: thread: nanotime: clock_gettime failed") # XXX +errno
            if not (0 <= ts.tv_sec and (0 <= ts.tv_nsec <= i1E9)):
                panic("pyxgo: thread: nanotime: clock_gettime -> invalid")
            if ts.tv_sec > (UINT64_MAX / i1E9 - 1):
                panic("pyxgo: thread: nanotime: clock_gettime -> overflow")
            return ts.tv_sec*i1E9 + ts.tv_nsec
    ELSE:
        uint64_t nanotime():
            cdef double t_s
            cdef PyExc exc
            with gil:
                pyexc_fetch(&exc)
                t_s, ok = _nanotime()
                pyexc_restore(exc)
            if not ok:
                panic("pyxgo: thread: nanotime: pytime.time failed")
            t_ns = t_s * 1E9
            if t_ns > UINT64_MAX:
                panic("pyxgo: thread: nanotime: time overflow")
            return <uint64_t>t_ns


    # XXX const
    _libgolang_runtime_ops thread_ops = _libgolang_runtime_ops(
            flags           = <_libgolang_runtime_flags>0,
            go              = go,
            sema_alloc      = sema_alloc,
            sema_free       = sema_free,
            sema_acquire    = sema_acquire,
            sema_release    = sema_release,
            nanosleep       = nanosleep,
            nanotime        = nanotime,
    )

from cpython cimport PyCapsule_New
libgolang_runtime_ops = PyCapsule_New(&thread_ops,
        "golang.runtime._runtime_thread.libgolang_runtime_ops", NULL)



# -*- coding: utf-8 -*-
# cython: language_level=2
# distutils: language=c++
#
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# small tests that verifies pyx-level sync API.
# the work of sync package itself is exercised thoroughly mostly in sync_test.py

from __future__ import print_function, absolute_import

from golang cimport go, chan, makechan, structZ, nil, panic, topyexc
from golang cimport sync, time

# C-level _sema + _mutex
# (not exposed in sync.pxd as it exposes only high-level API)
cdef extern from "golang/sync.h" namespace "golang::sync" nogil:
    """
    namespace golang {
    namespace sync {
        void _mutex_init(sync::Mutex *mu) {
            new (mu) sync::Mutex();
        }
        void _mutex_destroy(sync::Mutex *mu) {
            mu->~Mutex();
        }
    }}  // golang::sync::
    """
    struct _sema
    _sema *_makesema()
    void _semafree(_sema *)
    void _semaacquire(_sema *)
    void _semarelease(_sema *)

    void _mutex_init(sync.Mutex *)
    void _mutex_destroy(sync.Mutex *)

from libc.stdlib cimport calloc, free
from libc.stdio cimport printf


# test for semaphore wakeup, when wakeup is done not by the same thread which
# did the original acquire. This used to corrupt memory and deadlock on macOS
# due to CPython & PyPy runtime bugs:
#   https://bugs.python.org/issue38106
#   https://bitbucket.org/pypy/pypy/issues/3072
cdef nogil:
    struct WorkState:
        sync.Mutex mu       # protects vvv
        _sema      *sema    # T1 <- T2 wakeup; reallocated on every iteration
        bint stop           # T1 -> T2: request to stop
        chan[structZ] done  # T1 <- T2: stopped

    void _test_sema_wakeup_T2(void *_state):
        state = <WorkState*>_state
        cdef int i = 0, j
        cdef bint stop
        cdef double Wstart, now
        while 1:
            i += 1
            # wait till .sema != NULL and pop it
            Wstart = time.now()
            j = 0
            while 1:
                state.mu.lock()
                sema = state.sema
                if sema != NULL:
                    state.sema = NULL
                stop = state.stop
                state.mu.unlock()

                if stop:
                    state.done.close()
                    return
                if sema != NULL:
                    break

                now = time.now()
                if (now - Wstart) > 3*time.second:
                    printf("\nSTUCK on iteration #%d\n", i)
                    panic("STUCK")

                # workaround for "gevent" runtime: yield CPU so that T1 can run
                # XXX better yield always for "gevent", don't yield at all for "thread"
                j += 1
                if (j % 100) == 0:
                    time.sleep(0)

            # we have popped .sema from state. This means that peer is _likely_ waiting on it
            _semarelease(sema)  # either release or release + wakeup in-progress acquire


    void _test_sema_wakeup() except +topyexc:
        cdef WorkState *state = <WorkState *>calloc(1, sizeof(WorkState))
        if state == NULL:
            panic("malloc -> NULL")
        state.sema = NULL
        _mutex_init(&state.mu)
        state.stop = False
        state.done = makechan[structZ]()

        go(_test_sema_wakeup_T2, state)

        N = 100000
        cdef _sema *sema_prev = NULL
        for i in range(N):
            sema = _makesema()
            _semaacquire(sema)

            #printf("d(sema_prev, sema): %ld\n", <char*>sema - <char*>sema_prev)

            state.mu.lock()
            state.sema = sema
            state.mu.unlock()

            _semaacquire(sema)
            # _semarelease(sema) # (to free sema in released state)
            #                    # (gets stuck both with and without it)
            _semafree(sema)
            sema_prev = sema

        state.mu.lock()
        state.stop = True
        state.mu.unlock()

        state.done.recv()
        state.done = nil
        _mutex_destroy(&state.mu)
        free(state)

def test_sema_wakeup():
    with nogil:
        _test_sema_wakeup()


# verify Once
cdef nogil:
    int  _once_ncall = 0
    void _once_call():
        global _once_ncall
        _once_ncall += 1

cdef void _test_once() nogil except +topyexc:
    cdef sync.Once once
    if not (_once_ncall == 0):
        panic("once @0: ncall != 0")
    once.do(_once_call);
    if not (_once_ncall == 1):
        panic("once @1: ncall != 1")
    once.do(_once_call);
    if not (_once_ncall == 1):
        panic("once @2: ncall != 1")
    once.do(_once_call);
    if not (_once_ncall == 1):
        panic("once @3: ncall != 1")

def test_once():
    with nogil:
        _test_once()



#!/usr/bin/env python
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""This program is used to test traceback dump of chained exceptions.

When run it fails and Python should print the dump.
The dump is verified by test driver against golang_test_defer_excchain.txt .
"""

from __future__ import print_function, absolute_import

from golang import defer, func

def d1():
    raise RuntimeError("d1: aaa")
def d2():
    1/0
def d3():
    raise RuntimeError("d3: bbb")

@func
def main():
    defer(d3)
    defer(d2)
    defer(d1)
    raise RuntimeError("err")

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import select
from golang import time
from golang.golang_test import panics

# all timer tests operate in dt units
dt = 10*time.millisecond

# test_now verifies that time.now is correct.
def test_now():
    import time as stdtime
    assert stdtime is not time
    def tick(): # cpython 2.7 time.time uses max microsecond precision
        time.sleep(1*time.microsecond)
    t1 = stdtime.time() ;   tick()
    t2 = time.now()     ;   tick()
    t3 = stdtime.time() ;   tick()
    t4 = time.now()     ;   tick()
    assert t1 < t2
    assert t2 < t3
    assert t3 < t4


# test_{timer,ticker}_time verifie that Timer/Ticker, when they fire, send current time.
def test_timer_time():
    Tstart = time.now()
    t = time.Timer(3*dt)
    tnow = t.c.recv()
    Tend = time.now()

    assert type(tnow) is type(Tstart)
    assert Tstart < tnow < Tend
    assert ((tnow - Tstart) / dt) >= 3

def test_ticker_time():
    Tstart = time.now()
    tx = time.Ticker(3*dt)
    tnow = tx.c.recv()
    Tend = time.now()

    assert type(tnow) is type(Tstart)
    assert Tstart < tnow < Tend
    assert ((tnow - Tstart) / dt) >= 3


# test_timer verifies that Timer/Ticker fire as expected.
def test_timer():
    # start timers at x5, x7 and x11 intervals an verify that the timers fire
    # in expected sequence. The times when the timers fire do not overlap in
    # checked range because intervals are prime and chosen so that they start
    # overlapping only after 35 (=5·7).
    tv = [] # timer events
    Tstart = time.now()

    t23 = time.Timer(23*dt)
    t5  = time.Timer( 5*dt)

    def _():
        tv.append(7)
        t7f.reset(7*dt)
    t7f = time.Timer( 7*dt, f=_)

    tx11 = time.Ticker(11*dt)

    while 1:
        _, _rx = select(
            t23.c.recv,     # 0
            t5 .c.recv,     # 1
            t7f.c.recv,     # 2
            tx11.c.recv,    # 3
        )
        if _ == 0:
            tv.append(23)
            break
        if _ == 1:
            tv.append(5)
            t5.reset(5*dt)
        if _ == 2:
            assert False, "t7f sent to channel; must only call func"
        if _ == 3:
            tv.append(11)

    Tend = time.now()
    assert (Tend - Tstart) >= 23*dt
    assert tv == [        5,  7,     5, 11,       7, 5,             5, 7,11,23]
    #             1 2 3 4 5 6 7 8 9 10  11 12 13 14 15 16 17 18 19 20 21 22 23


# test_timer_misc, similarly to test_timer, verifies misc timer convenience functions.
def test_timer_misc():
    tv = []
    Tstart = time.now()

    c23 = time.after(23*dt)
    c5  = time.after( 5*dt)

    def _():
        tv.append(7)
        t7f.reset(7*dt)
    t7f = time.after_func(7*dt, _)

    cx11 = time.tick(11*dt)

    while 1:
        _, _rx = select(
            c23.recv,       # 0
            c5 .recv,       # 1
            t7f.c.recv,     # 2
            cx11.recv,      # 3
        )
        if _ == 0:
            tv.append(23)
            break
        if _ == 1:
            tv.append(5)
            # NOTE 5 does not rearm in this test because there is no way to
            # rearm timer create by time.after().
        if _ == 2:
            assert False, "t7f sent to channel; must only call func"
        if _ == 3:
            tv.append(11)

    Tend = time.now()
    assert (Tend - Tstart) >= 23*dt
    assert tv == [        5,  7,        11,       7,                   7,11,23]
    #             1 2 3 4 5 6 7 8 9 10  11 12 13 14 15 16 17 18 19 20 21 22 23


# test_timer_stop verifies that .stop() cancels Timer or Ticker.
def test_timer_stop():
    tv = []

    t10 = time.Timer (10*dt)
    t2  = time.Timer ( 2*dt)    # will fire and cancel t3, tx5
    t3  = time.Timer ( 3*dt)    # will be canceled
    tx5 = time.Ticker( 5*dt)    # will be canceled

    while 1:
        _, _rx = select(
            t10.c.recv,     # 0
            t2 .c.recv,     # 1
            t3 .c.recv,     # 2
            tx5.c.recv,     # 3
        )
        if _ == 0:
            tv.append(10)
            break
        if _ == 1:
            tv.append(2)
            t3.stop()
            tx5.stop()
        if _ == 2:
            tv.append(3)
        if _ == 3:
            tv.append(5)

    assert tv == [  2,              10]
    #             1 2 3 4 5 6 7 8 9 10


# test_timer_stop_drain verifies that Timer/Ticker .stop() drains timer channel.
def test_timer_stop_drain():
    t  = time.Timer (1*dt)
    tx = time.Ticker(1*dt)

    time.sleep(2*dt)
    assert len(t.c)  == 1
    assert len(tx.c) == 1

    assert t.stop() == False
    assert len(t.c) == 0

    tx.stop()
    assert len(tx.c) == 0


# test_timer_reset_armed verifies that .reset() panics if called on armed timer.
def test_timer_reset_armed():
    # reset while armed
    t = time.Timer(10*dt)
    with panics("Timer.reset: the timer is armed; must be stopped or expired"):
        t.reset(5*dt)



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""_runtime_gevent.pyx provides libgolang runtime based on gevent greenlets."""

from __future__ import print_function, absolute_import

# Gevent runtime uses gevent's greenlets and semaphores.
# When sema.acquire() blocks, gevent switches us from current to another greenlet.

IF not PYPY:
    from gevent._greenlet cimport Greenlet
    from gevent.__semaphore cimport Semaphore
    ctypedef Semaphore PYGSema
ELSE:
    # on pypy gevent does not compile greenlet.py and semaphore.py citing that
    # "there is no greenlet.h on pypy"
    from gevent.greenlet import Greenlet
    from gevent._semaphore import Semaphore
    ctypedef object PYGSema

from gevent import sleep as pygsleep

from libc.stdint cimport uint64_t
from cpython cimport Py_INCREF, Py_DECREF
from cython cimport final

from golang.runtime._libgolang cimport _libgolang_runtime_ops, _libgolang_sema, \
        STACK_DEAD_WHILE_PARKED, panic
from golang.runtime cimport _runtime_thread
from golang.runtime._runtime_pymisc cimport PyExc, pyexc_fetch, pyexc_restore


# _goviapy & _togo serve go
def _goviapy(_togo _ not None):
    with nogil:
        _.f(_.arg)

@final
cdef class _togo:
    cdef void (*f)(void *) nogil
    cdef void *arg


# internal functions that work under gil
cdef:
    # XXX better panic with pyexc object and detect that at recover side?

    bint _go(void (*f)(void *) nogil, void *arg):
        _ = _togo(); _.f = f; _.arg = arg
        g = Greenlet(_goviapy, _)
        g.start()
        return True

    _libgolang_sema* _sema_alloc():
        pygsema = Semaphore()
        Py_INCREF(pygsema)
        return <_libgolang_sema*>pygsema

    bint _sema_free(_libgolang_sema *gsema):
        pygsema = <PYGSema>gsema
        Py_DECREF(pygsema)
        return True

    bint _sema_acquire(_libgolang_sema *gsema):
        pygsema = <PYGSema>gsema
        pygsema.acquire()
        return True

    bint _sema_release(_libgolang_sema *gsema):
        pygsema = <PYGSema>gsema
        pygsema.release()
        return True

    bint _nanosleep(uint64_t dt):
        cdef double dt_s = dt * 1E-9
        pygsleep(dt_s)
        return True


# nogil runtime API
cdef nogil:

    void go(void (*f)(void *), void *arg):
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            ok = _go(f, arg)
            pyexc_restore(exc)
        if not ok:
            panic("pyxgo: gevent: go: failed")

    _libgolang_sema* sema_alloc():
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            sema = _sema_alloc()
            pyexc_restore(exc)
        return sema # libgolang checks for NULL return

    void sema_free(_libgolang_sema *gsema):
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            ok = _sema_free(gsema)
            pyexc_restore(exc)
        if not ok:
            panic("pyxgo: gevent: sema: free: failed")

    void sema_acquire(_libgolang_sema *gsema):
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            ok = _sema_acquire(gsema)
            pyexc_restore(exc)
        if not ok:
            panic("pyxgo: gevent: sema: acquire: failed")

    void sema_release(_libgolang_sema *gsema):
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            ok = _sema_release(gsema)
            pyexc_restore(exc)
        if not ok:
            panic("pyxgo: gevent: sema: release: failed")

    void nanosleep(uint64_t dt):
        cdef PyExc exc
        with gil:
            pyexc_fetch(&exc)
            ok = _nanosleep(dt)
            pyexc_restore(exc)
        if not ok:
            panic("pyxgo: gevent: sleep: failed")


    # XXX const
    _libgolang_runtime_ops gevent_ops = _libgolang_runtime_ops(
            # when greenlet is switched to another, its stack is copied to
            # heap, and stack of switched-to greenlet is copied back to C stack.
            flags           = STACK_DEAD_WHILE_PARKED,

            go              = go,
            sema_alloc      = sema_alloc,
            sema_free       = sema_free,
            sema_acquire    = sema_acquire,
            sema_release    = sema_release,
            nanosleep       = nanosleep,
            nanotime        = _runtime_thread.nanotime, # reuse from _runtime_thread
    )

from cpython cimport PyCapsule_New
libgolang_runtime_ops = PyCapsule_New(&gevent_ops,
        "golang.runtime._runtime_gevent.libgolang_runtime_ops", NULL)



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package pyx/runtime.pyx complements golang.pyx and provides support for
Python/Cython runtimes that can be used from nogil code.

 - `PyError` represents Python exception, that can be caught/reraised from
    nogil code, and is interoperated with libgolang `error`.
 - `PyFunc` represents Python function that can be called from nogil code.
"""

from golang  cimport error, _error, refptr, gobject, string
from cpython cimport PyObject


cdef extern from "golang/pyx/runtime.h" namespace "golang::pyx::runtime" nogil:
    const error ErrPyStopped

    cppclass _PyError (_error, gobject):
        string Error()

    cppclass PyError (refptr[_PyError]):
        # PyError.X = PyError->X in C++
        string Error "_ptr()->Error" ()

    error PyErr_Fetch()
    void  PyErr_ReRaise(PyError pyerr)

    cppclass PyFunc:
        __init__(PyObject *pyf)
        error operator() ()



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# Tests for pyx/runtime.pyx + libpyxruntime.

from __future__ import print_function, absolute_import

from golang cimport go, chan, makechan, structZ, select, panic
from golang cimport time
from golang.pyx cimport runtime

from cpython cimport PyObject
from libc.stdio cimport printf


ctypedef chan[structZ] chanZ
cdef structZ Z

# verify that there is no deadlock in libpyxruntime.pygil_ensure():
# it used to lock pyexitedMu -> GIL, which if also called with GIL initially
# held leads to deadlocks in AB BA scenario.
def test_pyfunc_vs_gil_deadlock():
    def f():
        time.sleep(0.001*time.second) # NOTE nogil sleep, _without_ releasing gil

    with nogil:
        _test_pyfunc_vs_gil_deadlock(runtime.PyFunc(<PyObject*>f))

cdef void _test_pyfunc_vs_gil_deadlock(runtime.PyFunc pyf) nogil:
    N = 100

    cdef chanZ ready = makechan[structZ]() # main <- spawned       "I'm ready"
    cdef chanZ start = makechan[structZ]() # main -> all spawned   "Start running"
    cdef chanZ done  = makechan[structZ]() # main <- spawned       "I'm finished"
    go(_runGM, N, pyf, ready, start, done)
    go(_runMG, N, pyf, ready, start, done)
    ready.recv(); ready.recv()
    start.close()

    cdef int ndone = 0
    timeoutq = time.after(5*time.second)
    while 1:
        _ = select([
            done.recvs(),       # 0
            timeoutq.recvs(),   # 1
        ])
        if _ == 0:
            ndone += 1
            if ndone == 2:
                break   # all ok
        if _ == 1:
            printf("\nSTUCK\n")
            panic("STUCK")

cdef nogil:

    # run in a loop locking GIL -> pyexitMu
    void _runGM(int n, runtime.PyFunc pyf, chanZ ready, chanZ start, chanZ done):
        ready.send(Z)
        start.recv()
        for i in range(n):
            #printf("GM %d\n", i)
            with gil:
                pyf()
        done.send(Z)

    # run in a loop locking pyexitMu -> GIL
    void _runMG(int n, runtime.PyFunc pyf, chanZ ready, chanZ start, chanZ done):
        ready.send(Z)
        start.recv()
        for i in range(n):
            #printf("MG %d\n", i)
            # already in nogil
            pyf()
        done.send(Z)



# -*- coding: utf-8 -*-
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package golang provides Go-like features for Python.

- `go` spawns lightweight thread.
- `chan` and `select` provide channels with Go semantic.
- `func` allows to define methods separate from class.
- `defer` allows to schedule a cleanup from the main control flow.
- `gimport` allows to import python modules by full path in a Go workspace.

See README for thorough overview.
See also package golang.pyx which provides similar functionality for Cython nogil.
"""

from __future__ import print_function, absolute_import

__version__ = "0.0.5"

__all__ = ['go', 'chan', 'select', 'default', 'nilchan', 'defer', 'panic', 'recover', 'func', 'gimport']

from golang._gopath import gimport  # make gimport available from golang
import inspect, sys
import decorator, six

from golang._golang import _pysys_exc_clear as _sys_exc_clear

# @func is a necessary decorator for functions for selected golang features to work.
#
# For example it is required by defer. Usage:
#
#   @func
#   def my_function(...):
#       ...
#
# @func can be also used to define methods separate from class, for example:
#
#   @func(MyClass)
#   def my_method(self, ...):
#       ...
def func(f):
    if inspect.isclass(f):
        fcall = inspect.currentframe().f_back   # caller's frame (where @func is used)
        return _meth(f, fcall)
    else:
        return _func(f)

# _meth serves @func(cls).
def _meth(cls, fcall):
    def deco(f):
        # wrap f with @_func, so that e.g. defer works automatically.
        f = _func(f)

        if isinstance(f, (staticmethod, classmethod)):
            func_name = f.__func__.__name__
        else:
            func_name = f.__name__
        setattr(cls, func_name, f)

        # if `@func(cls) def name` caller already has `name` set, don't override it
        missing = object()
        already = fcall.f_locals.get(func_name, missing)
        if already is not missing:
            return already

        # FIXME try to arrange so that python does not set anything on caller's
        # namespace[func_name]  (currently it sets that to implicitly returned None)

    return deco

# _func serves @func.
def _func(f):
    # @staticmethod & friends require special care:
    # unpack f first to original func and then repack back after wrapping.
    fclass = None
    if isinstance(f, (staticmethod, classmethod)):
        fclass = type(f)
        f = f.__func__

    def _(f, *argv, **kw):
        # run f under separate frame, where defer will register calls.
        __goframe__ = _GoFrame()
        with __goframe__:
            return f(*argv, **kw)

    # keep all f attributes, like __name__, __doc__, etc on _
    _ = decorator.decorate(f, _)

    # repack _ into e.g. @staticmethod if that was used on f.
    if fclass is not None:
        _ = fclass(_)

    return _

# _GoFrame serves __goframe__ that is setup by @func.
class _GoFrame:
    def __init__(self):
        self.deferv    = []     # defer registers funcs here
        self.recovered = False  # whether exception, if there was any, was recovered

        # py2: to-be next exception in exception chain (PEP 3134)
        if six.PY2:
            self.exc_ctx    = None  # exception context to chain new exception into
            self.exc_ctx_tb = None  # exc_tb we got when catching .exc_ctx.
                                    # we will set .exc_ctx.__traceback__ to this
                                    # if/when .exc_ctx will be chained into.
    def __enter__(self):
        pass

    # __exit__ simulates both except and finally.
    def __exit__(__goframe__, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            __goframe__.recovered = False

        # py2: simulate exception chaining (PEP 3134)
        if six.PY2:
            if exc_val is not None:
                # exc_val is current outer exception raised by e.g. earlier
                # defers; it can be itself chained.
                # .exc_ctx is current inner exception we saved before calling
                # code that raised exc_val. For example:
                #
                #   _GoFrame.__exit__:
                #       saves .exc_ctx      # .exc_ctx = A1
                #       with __goframe__:
                #           call other defer from .deferv
                #       __exit__(exc_val):  # exc_val = B3 (-> linked to B2 -> B1)
                #
                # the order in which exceptions were raised is: A1 B1 B2 B3
                # thus A1 is the context of B1, or in other words, .exc_ctx
                # should be linked to from tail of exc_val exception chain.
                exc_tail = exc_val
                while 1:
                    _ = getattr(exc_tail, '__context__', None)
                    if _ is None:
                        break
                    exc_tail = _
                exc_tail.__context__ = __goframe__.exc_ctx

                # make sure .__cause__ and .__suppress_context__ are always present
                if not hasattr(exc_val, '__cause__'):
                    exc_val.__cause__     = None
                if not hasattr(exc_val, '__suppress_context__'):
                    exc_val.__suppress_context__ = False

                # set .__traceback__ only for chained-to exceptions. top-level
                # raised exception must remain without __traceback__, because
                # if it was not yet caught, setting __traceback__ here early
                # will be wrong compared to what sys.exc_info() returns in
                # caller except block.
                if __goframe__.exc_ctx is not None:
                    __goframe__.exc_ctx.__traceback__ = __goframe__.exc_ctx_tb
                __goframe__.exc_ctx    = exc_val
                __goframe__.exc_ctx_tb = exc_tb

        if len(__goframe__.deferv) != 0:
            d = __goframe__.deferv.pop()

            # even if d panics - we have to call other defers
            with __goframe__:
                d()

        return __goframe__.recovered

# recover checks whether there is exception/panic currently being raised and returns it.
#
# If it was panic - it returns the argument that was passed to panic.
# If there is other exception - it returns the exception object.
#
# If there is no exception/panic, or the panic argument was None - recover returns None.
# Recover also returns None if it was not called by a deferred function directly.
def recover():
    fcall = inspect.currentframe().f_back   # caller's frame (deferred func)
    fgo   = fcall.f_back                    # caller's parent frame defined by _GoFrame.__exit__
    try:
        goframe = fgo.f_locals['__goframe__']
    except KeyError:
        # called not under go func/defer
        return None

    _, exc, _ = sys.exc_info()
    if exc is not None:
        goframe.recovered = True
        # recovered: clear current exception context
        _sys_exc_clear()
        if six.PY2:
            goframe.exc_ctx    = None
            goframe.exc_ctx_tb = None

    if type(exc) is _PanicError:
        exc = exc.args[0]
    return exc

# defer registers f to be called when caller function exits.
#
# It is similar to try/finally but does not force the cleanup part to be far
# away in the end.
def defer(f):
    fcall = inspect.currentframe().f_back   # caller's frame
    fgo   = fcall.f_back                    # caller's parent frame defined by @func
    try:
        goframe = fgo.f_locals['__goframe__']
    except KeyError:
        panic("function %s uses defer, but not @func" % fcall.f_code.co_name)

    goframe.deferv.append(f)


# py2: defer simulates exception chaining. Adjust traceback.print_exception()
# and default sys.excepthook so that, out of the box, dump of chained exceptions
# is printed with all details automatically.
if six.PY2:
    import traceback
    _tb_print_exception = traceback.print_exception
    def _print_exception(etype, value, tb, limit=None, file=None):
        if file is None:
            file = sys.stderr
        def emitf(msg):
            print(msg, file=file)
        def recursef(etype, value, tb):
            _print_exception(etype, value, tb, limit, file)

        _emit_exc_context(value, emitf, recursef)
        _tb_print_exception(etype, value, tb, limit, file)

    _tb_format_exception = traceback.format_exception
    def _format_exception(etype, value, tb, limit=None):
        l = []
        def emitf(msg):
            l.append(msg+"\n")
        def recursef(etype, value, tb):
            l.extend(_format_exception(etype, value, tb, limit))

        _emit_exc_context(value, emitf, recursef)
        l += _tb_format_exception(etype, value, tb, limit)
        return l

    # _emit_exc_context emits traceback for exc cause/context if any.
    #
    # emitf is used to emit raw text.
    # recursef is used to spawn processing on cause exception object.
    def _emit_exc_context(exc, emitf, recursef):
        ecause   = getattr(exc, '__cause__', None)
        econtext = getattr(exc, '__context__', None)
        if ecause is not None:
            recursef(type(ecause), ecause, getattr(ecause, '__traceback__', None))
            emitf("\nThe above exception was the direct cause of the following exception:\n")

        elif econtext is not None and not getattr(exc, '__suppress_context__', False):
            recursef(type(econtext), econtext, getattr(econtext, '__traceback__', None))
            emitf("\nDuring handling of the above exception, another exception occurred:\n")

    # patch traceback functions: in python2.7 all exception-related functions
    # in traceback module use either tb.print_exception() or tb.format_exception().
    # This way if we patch those two and someone uses e.g. tb.print_exc(),
    # it will print exception with cause/context included.
    traceback.print_exception  = _print_exception
    traceback.format_exception = _format_exception

    # adjust default sys.excepthook. Do this only if sys.excepthook was not already overridden.
    # Two cases are possible here:
    #   1) golang is imported in regular interpreter, possibly late in the process;
    #   2) golang is imported early as part of gpython startup.
    # For "2" when we get here the "pristine" precondition will be true, and so
    # we'll get to adjust sys.excepthook . For "1" if sys.excepthook is
    # pristine - it is safe to adjust. If sys.excepthook is not pristine - it
    # is not safe to adjust, because e.g. `import golang` was run from an
    # interactive IPython session and IPython already installed its own
    # sys.excepthook. We don't adjust sys.excepthook in such case, but we also
    # provide integration patches that add exception chaining support for
    # traceback dump functionality in popular third-party software.
    if sys.excepthook is sys.__excepthook__:
        sys.excepthook = traceback.print_exception

    # install pytest/ipython integration patches.
    # each patch is activated only when/if corresponding software is imported and actually used.
    import golang._patch.pytest_py2
    import golang._patch.ipython_py2


# ---- go + channels ----

from ._golang import    \
    pygo        as go,      \
    pychan      as chan,    \
    pyselect    as select,  \
    pydefault   as default, \
    pynilchan   as nilchan, \
    _PanicError,            \
    pypanic     as panic



# cython: language_level=2
# distutils: language=c++
#
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# Small program that uses a bit of golang.pyx nogil features, mainly to verify
# that external project can build against golang in pyx mode.

from golang cimport go, chan, makechan, topyexc
from libc.stdio cimport printf

cdef nogil:

    void worker(chan[int] ch, int i, int j):
        ch.send(i*j)

    void _main() except +topyexc:
        cdef chan[int] ch = makechan[int]()
        cdef int i
        for i in range(3):
            go(worker, ch, i, 4)

        for i in range(3):
            ch.recv()

        ch.close()
        #_, ok = ch.recv_() # TODO teach Cython to coerce pair[X,Y] -> (X,Y)
        ch.recv_()

        printf("test.pyx: OK\n")

def main():
    _main()



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package _patch contains patch infrastructure and patches that Pygolang
applies automatically."""

from __future__ import print_function, absolute_import

from peak.util import imports   # thanks PJE
import sys

# `@afterimport(modname) def f(mod)` arranges for f to be called after when, if
# at all, module modname will be imported.
#
# modname must be top-level module.
def afterimport(modname):
    if '.' in modname:
        raise AssertionError("BUG: modname has dot: %r" % (modname,))
    def _(f):
        def patchmod(mod):
            #print('patching %s ...' % (modname,))
            f()

        # XXX on pypy < 7.3 lazy-loading fails: https://bitbucket.org/pypy/pypy/issues/3099
        #     -> import & patch eagerly
        if 'PyPy' in sys.version and sys.pypy_version_info < (7,3):
            try:
                mod = __import__(modname)
            except ImportError:
                return  # module not available - nothing to patch
            patchmod(mod)
            return

        imports.whenImported(modname, patchmod)
        return f
    return _



# cython: language_level=2
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package sync mirrors and amends Go package sync.

 - `WorkGroup` allows to spawn group of goroutines working on a common task(*).
 - `Once` allows to execute an action only once.
 - `WaitGroup` allows to wait for a collection of tasks to finish.
 - `Sema`(*) and `Mutex` provide low-level synchronization.

See also https://golang.org/pkg/sync for Go sync package documentation.

(*) not provided in Go version.
"""

from golang cimport error, refptr
from golang cimport context

cdef extern from "golang/sync.h" namespace "golang::sync" nogil:
    cppclass Sema:
        Sema()
        void acquire()
        void release()

    cppclass Mutex:
        void lock()
        void unlock()

    cppclass Once:
        void do "do_" (...)     # ... = func<void()>

    cppclass WaitGroup:
        void done()
        void add(int delta)
        void wait()

    # WorkGroup
    cppclass _WorkGroup:
        void  go(...)                       # ... = func<error(context::Context)>
        error wait()

    cppclass WorkGroup (refptr[_WorkGroup]):
        # WorkGroup.X = WorkGroup->X in C++.
        void  go    "_ptr()->go"    (...)   # ... = func<error(context::Context)>
        error wait  "_ptr()->wait"  ()

    WorkGroup NewWorkGroup(context.Context ctx)



# pygolang | pythonic package setup
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from setuptools import find_packages
# setuptools has Library but this days it is not well supported and test for it
# has been killed https://github.com/pypa/setuptools/commit/654c26f78a30
# -> use setuptools_dso instead.
from setuptools_dso import DSO
from setuptools.command.install_scripts import install_scripts as _install_scripts
from setuptools.command.develop import develop as _develop
from distutils import sysconfig
from os.path import dirname, join
import sys, re

# read file content
def readfile(path):
    with open(path, 'r') as f:
        return f.read()

# reuse golang.pyx.build to build pygolang extensions.
# we have to be careful and inject synthetic golang package in order to be
# able to import golang.pyx.build without built/working golang.
trun = {}
exec(readfile('trun'), trun)
trun['ximport_empty_golangmod']()
from golang.pyx.build import setup, Extension as Ext


# grep searches text for pattern.
# return re.Match object or raises if pattern was not found.
def grep1(pattern, text):
    rex = re.compile(pattern, re.MULTILINE)
    m = rex.search(text)
    if m is None:
        raise RuntimeError('%r not found' % pattern)
    return m

# find our version
_ = readfile(join(dirname(__file__), 'golang/__init__.py'))
_ = grep1('^__version__ = "(.*)"$', _)
version = _.group(1)

# XInstallGPython customly installs bin/gpython.
#
# console_scripts generated by setuptools do lots of imports. However we need
# gevent.monkey.patch_all() to be done first - before all other imports. We
# could use plain scripts for gpython, however even for plain scripts
# setuptools wants to inject pkg_resources import for develop install, and
# pkg_resources does import lots of modules.
#
# -> generate the script via our custom install, but keep gpython listed as
# console_scripts entry point, so that pip knows to remove the file on develop
# uninstall.
#
# NOTE in some cases (see below e.g. about bdist_wheel) we accept for gpython
# to be generated not via XInstallGPython - because in those cases pkg_resources
# and entry points are not used - just plain `import gpython`.
class XInstallGPython:
    gpython_installed = 0

    # NOTE cannot override write_script, because base class - _install_scripts
    # or _develop, is old-style and super does not work with it.
    #def write_script(self, script_name, script, mode="t", blockers=()):
    #    script_name, script = self.transform_script(script_name, script)
    #    super(XInstallGPython, self).write_script(script_name, script, mode, blockers)

    # transform_script transform to-be installed script to override installed gpython content.
    #
    # (script_name, script) -> (script_name, script)
    def transform_script(self, script_name, script):
        # on windows setuptools installs 3 files:
        #   gpython-script.py
        #   gpython.exe
        #   gpython.exe.manifest
        # we want to override .py only.
        #
        # for-windows build could be cross - e.g. from linux via bdist_wininst -
        # -> we can't rely on os.name. Rely on just script name.
        if script_name in ('gpython', 'gpython-script.py'):
            script  = '#!%s\n' % sys.executable
            script += '\nfrom gpython import main; main()\n'
            self.gpython_installed += 1

        return script_name, script


# install_scripts is custom scripts installer that takes gpython into account.
class install_scripts(XInstallGPython, _install_scripts):
    def write_script(self, script_name, script, mode="t", blockers=()):
        script_name, script = self.transform_script(script_name, script)
        _install_scripts.write_script(self, script_name, script, mode, blockers)

    def run(self):
        _install_scripts.run(self)
        # bdist_wheel disables generation of scripts for entry-points[1]
        # and pip/setuptools regenerate them when installing the wheel[2].
        #
        #   [1] https://github.com/pypa/wheel/commit/0d7f398b
        #   [2] https://github.com/pypa/wheel/commit/9aaa6628
        #
        # since setup.py is not included into the wheel, we cannot control
        # entry-point installation when the wheel is installed. However,
        # console script generated when installing the wheel looks like:
        #
        #   #!/path/to/python
        #   # -*- coding: utf-8 -*-
        #   import re
        #   import sys
        #
        #   from gpython import main
        #
        #   if __name__ == '__main__':
        #       sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
        #       sys.exit(main())
        #
        # which does not import pkg_resources. Since we also double-check in
        # gpython itself that pkg_resources and other modules are not imported,
        # we are ok with this.
        if not self.no_ep:
            # regular install
            assert self.gpython_installed == 1
        else:
            # bdist_wheel
            assert self.gpython_installed == 0
            assert len(self.outfiles) == 0


# develop, similarly to install_scripts, is used to handle gpython in `pip install -e` mode.
class develop(XInstallGPython, _develop):
    def write_script(self, script_name, script, mode="t", blockers=()):
        script_name, script = self.transform_script(script_name, script)
        _develop.write_script(self, script_name, script, mode, blockers)

    def install_egg_scripts(self, dist):
        _develop.install_egg_scripts(self, dist)
        assert self.gpython_installed == 1


# requirements of packages under "golang." namespace
R = {
    'cmd.pybench':      {'pytest'},
    'pyx.build':        {'setuptools', 'wheel', 'cython', 'setuptools_dso >= 1.2'},
    'x.perf.benchlib':  {'numpy'},
}
# TODO generate `a.b -> a`, e.g. x.perf = join(x.perf.*); x = join(x.*)
Rall = set()
for pkg in R:
    Rall.update(R[pkg])
R['all'] = Rall

# extras_require <- R
extras_require = {}
for k in sorted(R.keys()):
    extras_require[k] = list(sorted(R[k]))


setup(
    name        = 'pygolang',
    version     = version,
    description = 'Go-like features for Python and Cython',
    long_description = '%s\n----\n\n%s' % (
                            readfile('README.rst'), readfile('CHANGELOG.rst')),
    long_description_content_type  = 'text/x-rst',
    url         = 'https://lab.nexedi.com/kirr/pygolang',
    project_urls= {
        'Bug Tracker':   'https://lab.nexedi.com/kirr/pygolang/issues',
        'Source Code':   'https://lab.nexedi.com/kirr/pygolang',
        'Documentation': 'https://pypi.org/project/pygolang',
    },
    license     = 'GPLv3+ with wide exception for Open-Source',
    author      = 'Kirill Smelkov',
    author_email= 'kirr@nexedi.com',

    keywords    = 'golang go channel goroutine concurrency GOPATH python import gpython gevent cython nogil GIL',

    packages    = find_packages(),

    x_dsos      = [DSO('golang.runtime.libgolang',
                        ['golang/runtime/libgolang.cpp',
                         'golang/context.cpp',
                         'golang/errors.cpp',
                         'golang/fmt.cpp',
                         'golang/strings.cpp',
                         'golang/sync.cpp',
                         'golang/time.cpp'],
                        depends = [
                            'golang/libgolang.h',
                            'golang/context.h',
                            'golang/cxx.h',
                            'golang/errors.h',
                            'golang/fmt.h',
                            'golang/strings.h',
                            'golang/sync.h',
                            'golang/time.h'],
                        include_dirs    = ['.', '3rdparty/include'],
                        define_macros   = [('BUILDING_LIBGOLANG', None)],
                        extra_compile_args = ['-std=gnu++11'], # not c++11 as linux/list.h uses typeof
                        soversion       = '0.1'),

                    DSO('golang.runtime.libpyxruntime',
                        ['golang/runtime/libpyxruntime.cpp'],
                        depends = ['golang/pyx/runtime.h'],
                        include_dirs    = ['.', sysconfig.get_python_inc()],
                        define_macros   = [('BUILDING_LIBPYXRUNTIME', None)],
                        extra_compile_args = ['-std=c++11'],
                        soversion       = '0.1',
                        dsos = ['golang.runtime.libgolang'])],

    ext_modules = [
                    Ext('golang._golang',
                        ['golang/_golang.pyx']),

                    Ext('golang.runtime._runtime_thread',
                        ['golang/runtime/_runtime_thread.pyx'],
                        language = "c"),

                    Ext('golang.runtime._runtime_gevent',
                        ['golang/runtime/_runtime_gevent.pyx'],
                        language = 'c'),

                    Ext('golang.pyx.runtime',
                        ['golang/pyx/runtime.pyx'],
                        dsos = ['golang.runtime.libpyxruntime']),

                    Ext('golang._golang_test',
                        ['golang/_golang_test.pyx',
                         'golang/runtime/libgolang_test_c.c',
                         'golang/runtime/libgolang_test.cpp']),

                    Ext('golang.pyx._runtime_test',
                        ['golang/pyx/_runtime_test.pyx'],
                        dsos = ['golang.runtime.libpyxruntime']),

                    Ext('golang._context',
                        ['golang/_context.pyx']),

                    Ext('golang._sync',
                        ['golang/_sync.pyx'],
                        dsos = ['golang.runtime.libpyxruntime'],
                        define_macros = [('_LIBGOLANG_SYNC_INTERNAL_API', None)]),
                    Ext('golang._sync_test',
                        ['golang/_sync_test.pyx']),

                    Ext('golang._time',
                        ['golang/_time.pyx'],
                        dsos = ['golang.runtime.libpyxruntime']),
                  ],
    include_package_data = True,

    install_requires = ['gevent', 'six', 'decorator', 'Importing;python_version<="2.7"'],
    extras_require   = extras_require,

    entry_points= {'console_scripts': [
                        # NOTE gpython is handled specially - see XInstallGPython.
                        'gpython  = gpython:main',

                        'py.bench = golang.cmd.pybench:main',
                      ]
                  },

    cmdclass    = {
        'install_scripts':  install_scripts,
        'develop':          develop,
    },

    classifiers = [_.strip() for _ in """\
        Development Status :: 3 - Alpha
        Intended Audience :: Developers
        Programming Language :: Python
        Programming Language :: Cython
        Programming Language :: Python :: 2
        Programming Language :: Python :: 2.7
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3.7
        Programming Language :: Python :: Implementation :: CPython
        Programming Language :: Python :: Implementation :: PyPy
        Topic :: Software Development :: Interpreters
        Topic :: Software Development :: Libraries :: Python Modules\
    """.splitlines()]
)



# -*- coding: utf-8 -*-
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package gcompat provides Go-compatibility layer for Python."""

from __future__ import print_function, absolute_import

from golang import strconv
import six

# qq is substitute for %q, which is missing in python.
#
# (python's automatic escape uses smartquotes quoting with either ' or ").
#
# like %s, %q automatically converts its argument to string.
def qq(obj):
    # make sure obj is text | bytes
    # py2: unicode | str
    # py3: str     | bytes
    if not isinstance(obj, (six.text_type, six.binary_type)):
        obj = str(obj)

    qobj = strconv.quote(obj)

    # `printf('%s', qq(obj))` should work. For this make sure qobj is always a
    # str - not bytes under py3 (if it was bytes it will print e.g. as b'...')
    if six.PY3 and isinstance(qobj, bytes):
        qobj = qobj.decode('UTF-8')

    return qobj



# -*- coding: utf-8 -*-
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang.strconv import quote, unquote, unquote_next
from golang.gcompat import qq

from six import int2byte as bchr, PY3
from six.moves import range as xrange
from pytest import raises

def byterange(start, stop):
    b = b""
    for i in xrange(start, stop):
        b += bchr(i)

    return b

def asstr(s):
    if PY3 and isinstance(s, bytes):
        s = s.decode('utf-8')
    return s

def test_quote():
    testv = (
        # in                quoted without leading/trailing "
        ('',                r""),

        (byterange(0,32),   br'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'),

        ('\'',              r"'"),
        ('"',               r"\""),
        ('ab c\ndef',       r"ab c\ndef"),
        ('a\'c\ndef',       r"a'c\ndef"),
        ('a\"c\ndef',       r"a\"c\ndef"),
        (u'a\"c\ndef',      u"a\\\"c\\ndef"),
        (b'a\"c\ndef',      br'a\"c\ndef'),
        ('привет\nмир',     r"привет\nмир"),
        (u'привет\nмир',    u"привет\\nмир"),

        # invalid utf-8
        (b"\xd0a",          br"\xd0a"),

        # non-printable utf-8
        (u"\u007f\u0080\u0081\u0082\u0083\u0084\u0085\u0086\u0087", u"\\x7f\\xc2\\x80\\xc2\\x81\\xc2\\x82\\xc2\\x83\\xc2\\x84\\xc2\\x85\\xc2\\x86\\xc2\\x87"),
    )

    for tin, tquoted in testv:
        # quote(in) == quoted
        # in = unquote(quoted)
        q = b'"' if isinstance(tquoted, bytes) else '"'
        tail = b'123' if isinstance(tquoted, bytes) else '123'
        tquoted = q + tquoted + q   # add lead/trail "

        assert quote(tin) == tquoted
        assert unquote(tquoted) == tin
        assert unquote_next(tquoted) == (tin, type(tin)())
        assert unquote_next(tquoted + tail) == (tin, tail)
        with raises(ValueError): unquote(tquoted + tail)

        # qq always gives str
        assert qq(tin) == asstr(tquoted)

        # also check how it works on complementary unicode/bytes input type
        if isinstance(tin, bytes):
            try:
                tin = tin.decode('utf-8')
            except UnicodeDecodeError:
                # some inputs are not valid UTF-8
                continue
            tquoted = tquoted.decode('utf-8')
            tail = tail.decode('utf-8')
        else:
            # tin was unicode
            tin = tin.encode('utf-8')
            tquoted = tquoted.encode('utf-8')
            tail = tail.encode('utf-8')

        assert quote(tin) == tquoted
        assert unquote(tquoted) == tin
        assert unquote_next(tquoted) == (tin, type(tin)())
        assert unquote_next(tquoted + tail) == (tin, tail)
        with raises(ValueError): unquote(tquoted + tail)

        # qq always gives str
        assert qq(tin) == asstr(tquoted)


def test_unquote_bad():
    testv = (
        # in            error
        ('x"zzz"',      'no starting "'),
        ('"zzz',        'no closing "'),
        ('"\\',         'unexpected EOL after \\'),
        ('"\\x',        'unexpected EOL after \\x'),
        ('"\\x0',       'unexpected EOL after \\x'),
        ('"\\z"',       'invalid escape \\z'),
    )

    for tin, err in testv:
        with raises(ValueError) as exc:
            unquote(tin)
        assert exc.value.args == (err,)



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from __future__ import print_function, absolute_import

from golang import nilchan, select, default
from golang import context, _context, time
from golang._context import _tctxAssertChildren as tctxAssertChildren
from golang.time_test import dt

# assertCtx asserts on state of _BaseCtx*
def assertCtx(ctx, children, deadline=None, err=None, done=False):
    assert isinstance(ctx, _context.PyContext)
    assert ctx.deadline() == deadline
    assert ctx.err() is err
    ctxdone = ctx.done()
    assert ready(ctxdone) == done
    tctxAssertChildren(ctx, children)
    for i in range(10): # repeated .done() returns the same pyobject
        assert ctx.done() is ctxdone

Z = set()   # empty set
C = context.canceled
D = context.deadlineExceeded
Y = True

bg = context.background()

# keys for context values
class Key:
    def __init__(self, key):
        self._key = key
    def __repr__(self):
        return "Key(%r)" % (self._key,)

    # __hash__ and __eq__ so that Key(x) == Key(x) to verify that keys are
    # compared by identity, _not_ equality.
    def __hash__(self):
        return hash(self._key)
    def __eq__(lhs, rhs):
        if not isinstance(rhs, Key):
            return False
        return lhs._key == rhs._key
kA      = Key("a")
assert kA == kA
assert not (kA != kA)
assert kA == Key("a")
assert kA != Key("b")
kHello  = Key("hello")
kHello2 = Key("hello")
assert kHello is not kHello2
assert kHello    ==  kHello2
kAbc    = Key("abc")
kBeta   = Key("beta")
kMir    = Key("мир")


# test_context exercises with_cancel / with_value and merge.
# deadlines are tested in test_deadline.
def test_context():
    assert bg.err()         is None
    assert bg.done()        == nilchan
    assert bg.deadline()    is None
    assert not ready(bg.done())
    assert bg.value(kHello) is None

    ctx1, cancel1 = context.with_cancel(bg)
    assert ctx1.done() != bg.done()
    assertCtx(ctx1,     Z)

    ctx11, cancel11 = context.with_cancel(ctx1)
    assert ctx11.done() != ctx1.done()
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    Z)

    vAlpha = Key("alpha")   # value object; Key is just reused as placeholder
    ctx111  = context.with_value(ctx11,  kHello, vAlpha)
    assert ctx111.done() == ctx11.done()
    assert ctx111.value(kHello)  is vAlpha  # original value object is returned
    assert ctx111.value(kHello2) is None    # _not_ vAlpha: keys are compared by identity
    assert ctx111.value(kAbc)    is None
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   Z)

    ctx1111 = context.with_value(ctx111, kBeta,  "gamma")
    assert ctx1111.done() == ctx11.done()
    assert ctx1111.value(kHello) is vAlpha
    assert ctx1111.value(kBeta)  == "gamma"
    assert ctx1111.value(kAbc)   is None
    assertCtx(ctx1,     {ctx11})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)


    ctx12 = context.with_value(ctx1, kHello, "world")
    assert ctx12.done() == ctx1.done()
    assert ctx12.value(kHello) == "world"
    assert ctx12.value(kAbc)   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    Z)

    ctx121, cancel121 = context.with_cancel(ctx12)
    assert ctx121.done() != ctx12.done()
    assert ctx121.value(kHello) == "world"
    assert ctx121.value(kAbc)   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   Z)

    ctx1211 = context.with_value(ctx121, kMir, "май")
    assert ctx1211.done() == ctx121.done()
    assert ctx1211.value(kHello) == "world"
    assert ctx1211.value(kMir)   == "май"
    assert ctx1211.value(kAbc)   is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  Z)
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   {ctx1211})
    assertCtx(ctx1211,  Z)

    ctxM, cancelM = context.merge(ctx1111, ctx1211)
    assert ctxM.done() != ctx1111.done()
    assert ctxM.done() != ctx1211.done()
    assert ctxM.value(kHello)   is vAlpha
    assert ctxM.value(kMir)     == "май"
    assert ctxM.value(kBeta)    == "gamma"
    assert ctxM.value(kAbc)     is None
    assertCtx(ctx1,     {ctx11, ctx12})
    assertCtx(ctx11,    {ctx111})
    assertCtx(ctx111,   {ctx1111})
    assertCtx(ctx1111,  {ctxM})
    assertCtx(ctx12,    {ctx121})
    assertCtx(ctx121,   {ctx1211})
    assertCtx(ctx1211,  {ctxM})
    assertCtx(ctxM,     Z)


    for _ in range(2):
        cancel11()
        assertCtx(ctx1,     {ctx12})
        assertCtx(ctx11,    Z, err=C, done=Y)
        assertCtx(ctx111,   Z, err=C, done=Y)
        assertCtx(ctx1111,  Z, err=C, done=Y)
        assertCtx(ctx12,    {ctx121})
        assertCtx(ctx121,   {ctx1211})
        assertCtx(ctx1211,  Z)
        assertCtx(ctxM,     Z, err=C, done=Y)

    for _ in range(2):
        cancel1()
        assertCtx(ctx1,     Z, err=C, done=Y)
        assertCtx(ctx11,    Z, err=C, done=Y)
        assertCtx(ctx111,   Z, err=C, done=Y)
        assertCtx(ctx1111,  Z, err=C, done=Y)
        assertCtx(ctx12,    Z, err=C, done=Y)
        assertCtx(ctx121,   Z, err=C, done=Y)
        assertCtx(ctx1211,  Z, err=C, done=Y)
        assertCtx(ctxM,     Z, err=C, done=Y)


# test_deadline exercises deadline-related context functionality.
def test_deadline():
    t0 = time.now()
    d1 = t0 + 10*dt
    d2 = t0 + 20*dt
    d3 = t0 + 30*dt

    ctx1, cancel1 = context.with_deadline(bg, d2)
    assert ctx1.done() != bg.done()
    assertCtx(ctx1, Z, deadline=d2)

    ctx11 = context.with_value(ctx1, kA, "b")
    assert ctx11.done() == ctx1.done()
    assert ctx11.value(kA) == "b"
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    Z,              deadline=d2)

    ctx111, cancel111 = context.with_cancel(ctx11)
    assert ctx111.done() != ctx11.done
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   Z,              deadline=d2)

    ctx1111, cancel1111 = context.with_deadline(ctx111, d3) # NOTE deadline > parent
    assert ctx1111.done() != ctx111.done()
    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)    # NOTE not d3

    ctx12, cancel12 = context.with_deadline(ctx1, d1)
    assert ctx12.done() != ctx1.done()
    assertCtx(ctx1,     {ctx11, ctx12}, deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)
    assertCtx(ctx12,    Z,              deadline=d1)

    ctxM, cancelM = context.merge(ctx1111, ctx12)
    assert ctxM.done() != ctx1111.done()
    assert ctxM.done() != ctx12.done()
    assert ctxM.value(kA) == "b"
    assertCtx(ctx1,     {ctx11, ctx12}, deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  {ctxM},         deadline=d2)
    assertCtx(ctx12,    {ctxM},         deadline=d1)
    assertCtx(ctxM,     Z,              deadline=d1)

    time.sleep(11*dt)

    assertCtx(ctx1,     {ctx11},        deadline=d2)
    assertCtx(ctx11,    {ctx111},       deadline=d2)
    assertCtx(ctx111,   {ctx1111},      deadline=d2)
    assertCtx(ctx1111,  Z,              deadline=d2)
    assertCtx(ctx12,    Z,              deadline=d1, err=D, done=Y)
    assertCtx(ctxM,     Z,              deadline=d1, err=D, done=Y)

    # explicit cancel first -> err=canceled instead of deadlineExceeded
    for i in range(2):
        cancel1()
        assertCtx(ctx1,     Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx11,    Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx111,   Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx1111,  Z,          deadline=d2, err=C, done=Y)
        assertCtx(ctx12,    Z,          deadline=d1, err=D, done=Y)
        assertCtx(ctxM,     Z,          deadline=d1, err=D, done=Y)


    # with_timeout
    ctx, cancel = context.with_timeout(bg, 10*dt)
    assert ctx.done() != bg.done()
    d = ctx.deadline()
    assert abs(d - (time.now() + 10*dt)) < 1*dt
    assertCtx(ctx,  Z,  deadline=d)

    time.sleep(11*dt)
    assertCtx(ctx,  Z,  deadline=d, err=D, done=Y)


# ---- misc ----

# _ready returns whether channel ch is ready.
def ready(ch):
    _, _rx = select(
            ch.recv,    # 0
            default,    # 1
    )
    if _ == 0:
        return True
    if _ == 1:
        return False



# -*- coding: utf-8 -*-
# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package context mirrors and amends Go package context.

 - `Context` represents operational context that carries deadline, cancellation
   signal and immutable context-local key -> value dict.
 - `background` returns empty context that is never canceled.
 - `with_cancel` creates new context that can be canceled on its own.
 - `with_deadline` creates new context with deadline.
 - `with_timeout` creates new context with timeout.
 - `with_value` creates new context with attached key=value.
 - `merge` creates new context from 2 parents(*).

See also https://golang.org/pkg/context for Go context package documentation.
See also https://blog.golang.org/context for overview.

(*) not provided in Go version.
"""

from __future__ import print_function, absolute_import

from golang._context import \
    PyContext               as Context,             \
    pybackground            as background,          \
    pycanceled              as canceled,            \
    pydeadlineExceeded      as deadlineExceeded,    \
    pywith_cancel           as with_cancel,         \
    pywith_value            as with_value,          \
    pywith_deadline         as with_deadline,       \
    pywith_timeout          as with_timeout,        \
    pymerge                 as merge
